package ch.epfl.ppo.collections;

import java.util.Iterator;
import java.util.NoSuchElementException;

public final class TreeSet<E extends Comparable<E>> implements Set<E> {
    private int size = 0;
    private Node<E> root = null;

    @Override
    public boolean isEmpty() {
        return size == 0;
    }

    @Override
    public int size() {
        return size;
    }

    @Override
    public void add(E newElem) {
        if (root == null) {
            root = new Node<>(newElem);
        } else {
            Node<E> n = nodeOrParentFor(newElem);
            int cmp = newElem.compareTo(n.elem);
            if (cmp < 0) {
                assert n.smaller == null;
                n.smaller = new Node<>(n, newElem);
            } else if (cmp > 0) {
                assert n.greater == null;
                n.greater = new Node<>(n, newElem);
            } else
                return;
        }
        size += 1;
    }

    @Override
    public void remove(E elem) {
        Node<E> n = nodeFor(elem);
        if (n != null)
            removeNode(n);
    }

    @Override
    public boolean contains(E elem) {
        return nodeFor(elem) != null;
    }

    @Override
    public Iterator<E> iterator() {
        return new Iterator<E>() {
            private Node<E> next, last;

            {
                last = null;
                next = root;
                while (next != null && next.smaller != null)
                    next = next.smaller;
            }

            @Override
            public boolean hasNext() {
                return next != null;
            }

            @Override
            public E next() {
                if (! hasNext())
                    throw new NoSuchElementException();
                E elem = next.elem;
                last = next;
                next = successor(next);
                return elem;
            }

            @Override
            public void remove() {
                if (last == null)
                    throw new IllegalStateException();
                removeNode(last);
                last = null;
            }
        };
    }

    private Node<E> nodeOrParentFor(E elem) {
        Node<E> n = root, p = null;
        while (n != null) {
            int cmp = elem.compareTo(n.elem);
            p = n;
            if (cmp < 0) n = n.smaller;
            else if (cmp > 0) n = n.greater;
            else return n;
        }
        return p;
    }

    private Node<E> nodeFor(E elem) {
        Node<E> nodeOrParent = nodeOrParentFor(elem);
        return nodeOrParent != null && nodeOrParent.elem.equals(elem)
                ? nodeOrParent
                        : null;
    }

    private static <E> Node<E> successor(Node<E> n) {
        if (n.greater != null) {
            Node<E> s = n.greater;
            while (s.smaller != null)
                s = s.smaller;
            return s;
        } else {
            Node<E> s = n.parent;
            Node<E> c = n;
            while (s != null && s.greater == c) {
                c = s;
                s = s.parent;
            }
            return s;
        }
    }

    private void removeNode(Node<E> n) {
        if (n.smaller != null && n.greater != null) {
            Node<E> s = successor(n);
            n.elem = s.elem;
            n = s;
        }
        assert n.smaller == null || n.greater == null;
        Node<E> maybeChild = (n.smaller != null ? n.smaller : n.greater);
        if (maybeChild != null)
            maybeChild.parent = n.parent;
        if (n.parent == null)
            root = maybeChild;
        else if (n == n.parent.smaller)
            n.parent.smaller = maybeChild;
        else
            n.parent.greater = maybeChild;
        size -= 1;
    }

    private final static class Node<E> {
        public E elem;
        public Node<E> parent;
        public Node<E> smaller, greater;

        public Node(Node<E> parent, Node<E> smaller, E elem, Node<E> greater) {
            this.parent = parent;
            this.smaller = smaller;
            this.elem = elem;
            this.greater = greater;
        }

        public Node(Node<E> parent, E elem) {
            this(parent, null, elem, null);
        }

        public Node(E value) {
            this(null, null, value, null);
        }
    }
}
